#!/usr/bin/env python
from sklearn.metrics import f1_score
from sklearn.preprocessing import LabelEncoder
# Scoring program for the AutoML challenge
# Isabelle Guyon and Arthur Pesah, ChaLearn, August 2014-November 2016

# ALL INFORMATION, SOFTWARE, DOCUMENTATION, AND DATA ARE PROVIDED "AS-IS". 
# ISABELLE GUYON, CHALEARN, AND/OR OTHER ORGANIZERS OR CODE AUTHORS DISCLAIM
# ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR ANY PARTICULAR PURPOSE, AND THE
# WARRANTY OF NON-INFRINGEMENT OF ANY THIRD PARTY'S INTELLECTUAL PROPERTY RIGHTS. 
# IN NO EVENT SHALL ISABELLE GUYON AND/OR OTHER ORGANIZERS BE LIABLE FOR ANY SPECIAL, 
# INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF SOFTWARE, DOCUMENTS, MATERIALS, 
# PUBLICATIONS, OR INFORMATION MADE AVAILABLE FOR THE CHALLENGE. 
from functools import reduce
# Some libraries and options
import os
from sys import argv
from glob import glob
import numpy as np
import scipy as sp
# import libscores
# import yaml
# from libscores import ls, filesep, mkdir, read_array, compute_all_scores, write_scores
# from libscores import ls, filesep, mkdir, read_array, compute_all_scores, write_scores

if (os.name == "nt"):
    filesep = '\\'
else:
    filesep = '/'

# Default I/O directories:
root_dir = "/home/lucas/EMNLP_codalab/iris/"
default_input_dir = root_dir + "sample_input"
default_output_dir = root_dir + "scoring_output"

# Debug flag 0: no debug, 1: show all scores, 2: also show version amd listing of dir
debug_mode = 1

# Constant used for a missing score
missing_score = -0.999999

# Version number
scoring_version = 1.0

def mkdir(d):
    if not os.path.exists(d):
        os.makedirs(d)

def _HERE(*args):
    h = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(h, *args)

def binarize_predictions(array, task='binary.classification'):
    ''' Turn predictions into decisions {0,1} by selecting the class with largest 
    score for multiclass problems and thresholding at 0.5 for other cases.'''
    # add a very small random value as tie breaker (a bit bad because this changes the score every time)
    # so to make sure we get the same result every time, we seed it    
    # eps = 1e-15
    # np.random.seed(sum(array.shape))
    # array = array + eps*np.random.rand(array.shape[0],array.shape[1])
    bin_array = np.zeros(array.shape)
    if (task != 'multiclass.classification') or (array.shape[1] == 1):
        bin_array[array >= 0.5] = 1
    else:
        sample_num = array.shape[0]
        for i in range(sample_num):
            j = np.argmax(array[i, :])
            bin_array[i, j] = 1
    return bin_array

def acc_stat(solution, prediction):
    ''' Return accuracy statistics TN, FP, TP, FN
     Assumes that solution and prediction are binary 0/1 vectors.'''
    # This uses floats so the results are floats
    TN = sum(np.multiply((1 - solution), (1 - prediction)))
    FN = sum(np.multiply(solution, (1 - prediction)))
    TP = sum(np.multiply(solution, prediction))
    FP = sum(np.multiply((1 - solution), prediction))
    # print "TN =",TN
    # print "FP =",FP
    # print "TP =",TP
    # print "FN =",FN
    return (TN, FP, TP, FN)

def mvmean(R, axis=0):
    ''' Moving average to avoid rounding errors. A bit slow, but...
    Computes the mean along the given axis, except if this is a vector, in which case the mean is returned.
    Does NOT flatten.'''
    if len(R.shape) == 0: return R
    average = lambda x: reduce(lambda i, j: (0, (j[0] / (j[0] + 1.)) * i[1] + (1. / (j[0] + 1)) * j[1]), enumerate(x))[
        1]
    R = np.array(R)
    if len(R.shape) == 1: return average(R)
    if axis == 1:
        return np.array(map(average, R))
    else:
        return np.array(map(average, R.transpose()))

def f1_metric(solution, prediction, task='binary.classification'):
    ''' Compute the normalized f1 measure. The binarization differs 
        for the multi-label and multi-class case. 
        A non-weighted average over classes is taken.
        The score is normalized.'''
    label_num = solution.shape[1]
    score = np.zeros(label_num)
    bin_prediction = binarize_predictions(prediction, task)
    [tn, fp, tp, fn] = acc_stat(solution, bin_prediction)
    # Bounding to avoid division by 0
    eps = 1e-15
    true_pos_num = sp.maximum(eps, tp + fn)
    found_pos_num = sp.maximum(eps, tp + fp)
    tp = sp.maximum(eps, tp)
    tpr = tp / true_pos_num  # true positive rate (recall)
    ppv = tp / found_pos_num  # positive predictive value (precision)
    arithmetic_mean = 0.5 * sp.maximum(eps, tpr + ppv)
    # Harmonic mean:
    f1 = tpr * ppv / arithmetic_mean
    # Average over all classes
    f1 = mvmean(f1)
    # Normalize: 0 for random, 1 for perfect
    if (task != 'multiclass.classification') or (label_num == 1):
        # How to choose the "base_f1"?
        # For the binary/multilabel classification case, one may want to predict all 1.
        # In that case tpr = 1 and ppv = frac_pos. f1 = 2 * frac_pos / (1+frac_pos)
        #     frac_pos = mvmean(solution.ravel())
        #     base_f1 = 2 * frac_pos / (1+frac_pos)
        # or predict random values with probability 0.5, in which case
        #     base_f1 = 0.5
        # the first solution is better only if frac_pos > 1/3.
        # The solution in which we predict according to the class prior frac_pos gives
        # f1 = tpr = ppv = frac_pos, which is worse than 0.5 if frac_pos<0.5
        # So, because the f1 score is used if frac_pos is small (typically <0.1)
        # the best is to assume that base_f1=0.5
        base_f1 = 0.5
    # For the multiclass case, this is not possible (though it does not make much sense to
    # use f1 for multiclass problems), so the best would be to assign values at random to get 
    # tpr=ppv=frac_pos, where frac_pos=1/label_num
    else:
        base_f1 = 1. / label_num
    score = (f1 - base_f1) / sp.maximum(eps, (1 - base_f1))
    return score

def _load_scoring_function():
    dict_metrics = {'f1_metric':f1_metric}

    return "f1_metric", dict_metrics['f1_metric']
    # with open(_HERE('metric.txt'), 'r') as f:
    #     metric_name = f.readline().strip()
    #     return metric_name, getattr(libscores, metric_name)

def ls(filename):
    return sorted(glob(filename))

def write_scores(fp, scores):
    ''' Write scores to file opened under file pointer fp'''
    for key in scores.keys():
        fp.write("%s --> %s\n" % (key, scores[key]))
        print(key + " --> " + str(scores[key]))

# def read_array(filename):
#     ''' Read array and convert to 2d np arrays '''
#     array = np.genfromtxt(filename, dtype=float)
#     if len(array.shape) == 1:
#         array = array.reshape(-1, 1)
#     return array

def read_array(filename):
    array = open(filename,'r').readlines()
    array = [x.rstrip() for x in array]
    return array


def get_label_names(filename):
    with open(filename) as f:
        content = f.readlines()
    content = [x.rstrip() for x in content]
    labels_name = []
    for line in content:
        labels_name += [line]
    return labels_name

# =============================== MAIN ========================================
# exit()
if __name__ == "__main__":

    #### INPUT/OUTPUT: Get input and output directory names
    if len(argv) == 1:  # Use the default input and output directories if no arguments are provided
        input_dir = default_input_dir
        output_dir = default_output_dir
    else:
        input_dir = argv[1]
        output_dir = argv[2]
        # Create the output directory, if it does not already exist and open output files
    mkdir(output_dir)
    score_file = open(os.path.join(output_dir, 'scores.txt'), 'wb')
    html_file = open(os.path.join(output_dir, 'scores.html'), 'wb')

    # Get the metric
    # metric_name, scoring_function = _load_scoring_function()

    # Get all the solution files from the solution directory
    # exit()
    solution_names = sorted(ls(os.path.join(input_dir, 'ref', '*.solution')))
    labels_name = get_label_names(os.path.join(input_dir, 'ref', 'labels.txt'))
    labelencoder = LabelEncoder()
    labelencoder.fit(labels_name)
    # print(labelencoder.classes_)
    # exit()
    # Loop over files in solution directory and search for predictions with extension .predict having the same basename
    for i, solution_file in enumerate(solution_names):
        set_num = i + 1  # 1-indexed
        score_name = 'set%s_score' % set_num

        # Extract the dataset name from the file name
        basename = solution_file[-solution_file[::-1].index(filesep):-solution_file[::-1].index('.') - 1]
        
        try:
            # Get the last prediction from the res subdirectory (must end with '.predict')
            
            predict_file = ls(os.path.join(input_dir, 'res', basename + '*.predict'))[-1]
            
            if (predict_file == []): raise IOError('Missing prediction file {}'.format(basename))
            
            predict_name = predict_file[-predict_file[::-1].index(filesep):-predict_file[::-1].index('.') - 1]
            # Read the solution and prediction values into numpy arrays

            solution = read_array(solution_file)
            
            prediction = read_array(predict_file)
            solution = labelencoder.transform(np.array(solution))
            prediction = labelencoder.transform(np.array(prediction))
            
            if (solution.shape != prediction.shape): raise ValueError(
                "Bad prediction shape {}".format(prediction.shape))

            try:
                # Compute the score prescribed by the metric file 
                # score = scoring_function(solution, prediction)
                score = f1_score(solution, prediction, average='macro')
                score2 = f1_score(solution, prediction, average='micro')
                  
                # f1_score(y_true, y_pred, average='micro')  
                print(
                    "======= Set %d" % set_num + " (" + predict_name.capitalize() + "): score(" + score_name + ")=%0.12f =======" % score)
                print(
                    "======= Set %d" % 2 + " (" + predict_name.capitalize() + "): score(" + 'set2_score' + ")=%0.12f =======" % score2)
                html_file.write((
                    "======= Set %d" % set_num + " (" + predict_name.capitalize() + "): score(" + score_name + ")=%0.12f =======\n" % score).encode())
                html_file.write((
                    "======= Set %d" % 2 + " (" + predict_name.capitalize() + "): score(" + 'set2_score' + ")=%0.12f =======\n" % score2).encode())
            except:
                raise Exception('Error in calculation of the specific score of the task')
            # if debug_mode > 0:
            #     scores = compute_all_scores(solution, prediction)
            #     write_scores(html_file, scores)

        except Exception as inst:
            score = missing_score
            print(
                "======= Set %d" % set_num + " (" + basename.capitalize() + "): score(" + score_name + ")=ERROR =======")
            print(
                "======= Set %d" % 2 + " (" + basename.capitalize() + "): score(" + 'set2_score' + ")=ERROR =======")
            html_file.write(
                "======= Set %d" % set_num + " (" + basename.capitalize() + "): score(" + score_name + ")=ERROR =======\n" + "inst:" + str(inst))
            html_file.write(
                "======= Set %d" % 2 + " (" + basename.capitalize() + "): score(" + 'set2_score' + ")=ERROR =======\n" + "inst:" + str(inst))


        # Write score corresponding to selected task and metric to the output file
        score_file.write((score_name + ": %0.12f\n" % score).encode())
        score_file.write(('set2_score' + ": %0.12f\n" % score2).encode())
    # End loop for solution_file in solution_names

    # Read the execution time and add it to the scores:
    try:
        metadata = yaml.load(open(os.path.join(input_dir, 'res', 'metadata'), 'r'))
        score_file.write(("Duration: %0.6f\n" % metadata['elapsedTime']).encode())
    except:
        score_file.write(("Duration: 0\n").encode())

        html_file.close()
    score_file.close()

    # Lots of debug stuff
    if debug_mode > 1:
        swrite('\n*** SCORING PROGRAM: PLATFORM SPECIFICATIONS ***\n\n')
        show_platform()
        show_io(input_dir, output_dir)
        show_version(scoring_version)

        # exit(0)
